@@ -88,7 +88,6 @@ namespace GPBoost {
 		/*!
 		* \brief Virtual function that calculates the covariance matrix Z*Sigma*Z^T
 		* \return Covariance matrix Z*Sigma*Z^T of this component
-		*   Note that since sigma_ is saved (since it is used in GetZSigmaZt and GetZSigmaZtGrad) we return a pointer and do not write on an input parameter in order to avoid copying
 		*/
 		virtual std::shared_ptr<T_mat> GetZSigmaZt() const = 0;
 
@@ -624,7 +623,37 @@ namespace GPBoost {
 					}
 				}
 			}//end not is_rand_coef_
-		}
+		}//end CalcInsertZtilde
+
+		/*!
+		* \brief Calculate and add unconditional predictive variances only for new groups that do not appear in training data
+		* \param[out] pred_uncond_var Array of unconditional predictive variances to which the variance of this component is added
+		* \param num_data_pred Number of prediction points
+		* \param rand_coef_data_pred Covariate data for varying coefficients
+		* \param group_data_pred Group data for predictions
+		*/
+		void AddPredUncondVarNewGroups(double* pred_uncond_var,
+			int num_data_pred,
+			const double* const rand_coef_data_pred,
+			const std::vector<re_group_t>& group_data_pred) const {
+			CHECK(num_data_pred == (int)group_data_pred.size());
+			if (this->is_rand_coef_) {
+#pragma omp parallel for schedule(static)
+				for (int i = 0; i < num_data_pred; ++i) {
+					if (map_group_label_index_->find(group_data_pred[i]) == map_group_label_index_->end()) {//Group level 'group_data_pred[i]' does not exist in observed data
+						pred_uncond_var[i] += this->cov_pars_[0] * rand_coef_data_pred[i] * rand_coef_data_pred[i];
+					}
+				}
+			}//end is_rand_coef_
+			else {//not is_rand_coef_
+#pragma omp parallel for schedule(static)
+				for (int i = 0; i < num_data_pred; ++i) {
+					if (map_group_label_index_->find(group_data_pred[i]) == map_group_label_index_->end()) {//Group level 'group_data_pred[i]' does not exist in observed data
+						pred_uncond_var[i] += this->cov_pars_[0];
+					}
+				}
+			}//end not is_rand_coef_
+		}//end AddPredUncondVarNewGroups
 
 		data_size_t GetNumUniqueREs() const override {
 			return(num_group_);
@@ -929,7 +958,8 @@ namespace GPBoost {
 			is_cross_covariance_IP_ = true;
 			apply_tapering_ = apply_tapering;
 			apply_tapering_manually_ = apply_tapering_manually;
-			cov_function_ = std::shared_ptr<CovFunction<T_mat>>(new CovFunction<T_mat>(cov_fct, shape, taper_range, taper_shape, taper_mu, apply_tapering, (int)coords.cols(), true));
+			bool save_distances = false;
+			cov_function_ = std::shared_ptr<CovFunction<T_mat>>(new CovFunction<T_mat>(cov_fct, shape, taper_range, taper_shape, taper_mu, apply_tapering, (int)coords.cols(), save_distances));
 			has_compact_cov_fct_ = (COMPACT_SUPPORT_COVS_.find(cov_function_->cov_fct_type_) != COMPACT_SUPPORT_COVS_.end()) || apply_tapering_;
 			this->num_cov_par_ = cov_function_->num_cov_par_;
 			coords_ind_point_ = coords_ind_point;
@@ -954,7 +984,7 @@ namespace GPBoost {
 				coords_ = coords;
 			}
 			num_random_effects_ = (data_size_t)coords_.rows();
-			if (cov_function_->IsIsotropic() || apply_tapering_ || apply_tapering_manually_) {
+			if ((save_distances && cov_function_->IsIsotropic()) || apply_tapering_ || apply_tapering_manually_) {
 				//Calculate distances
 				T_mat dist;
 				if (has_compact_cov_fct_) {//compactly suported covariance
@@ -1201,6 +1231,10 @@ namespace GPBoost {
 			(*cov_function_).template MultiplyWendlandCorrelationTaper<T_mat>(dist, sigma, false);
 		}
 
+		const T_mat* GetSigmaPtr() const {
+			return(&sigma_);
+		}
+
 		/*!
 		* \brief Calculate covariance matrix
 		* \return Covariance matrix Z*Sigma*Z^T of this component
