@@ -159,7 +159,6 @@ public:
             // Test the curvature condition
             if (std::abs(dg) <= test_curv)
             {
-                f(x, grad, false, true);//calculate gradient
                 return;  // Case (4)
             }
                 
@@ -192,7 +191,6 @@ public:
                 // But we need to move {x, grad}_lo back before returning
                 x.swap(x_lo);
                 grad.swap(grad_lo);
-                f(x, grad, false, true);//calculate gradient
                 return;
             }
 
@@ -272,6 +270,7 @@ public:
                 if (step_lo <= Scalar(0))
                 {
                     x.noalias() = xp;
+                    f.ResetProfiledOutVariablesToLag1();
                     fx = fx_init;
                     step = 0.;
                     Log::REDebug("GPModel lbfgs: the line search routine reached the maximum number of iterations");
@@ -286,7 +285,6 @@ public:
                     x.swap(x_lo);
                     grad.swap(grad_lo);
                 }
-                f(x, grad, false, true);//calculate gradient
                 return;
             }
         }
